from datetime import datetime
from typing import Literal
from zoneinfo import ZoneInfo

from pydantic import BaseConfig, Extra, validator

from weaverbird.pipeline.steps.utils.base import BaseStep, to_camelcase
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class TextStep(BaseStep):
    name: Literal["text"] = "text"
    text: datetime | int | float | bool | str
    new_column: ColumnName

    class Config(BaseConfig):
        allow_population_by_field_name = True
        extra = Extra.forbid
        alias_generator = to_camelcase
        smart_union = True

    @validator("text")
    def _text_validator(cls, value):
        if isinstance(value, datetime) and value.tzinfo is not None:
            return value.astimezone(ZoneInfo("UTC")).replace(tzinfo=None)
        return value


class TextStepWithVariable(TextStep, StepWithVariablesMixin):
    ...



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DeleteStep


def translate_delete(step: DeleteStep) -> list[MongoStep]:
    return [{"$project": {c: 0 for c in step.columns}}]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import WaterfallStep


def translate_waterfall(step: WaterfallStep) -> list[MongoStep]:
    concat_mongo = {}
    facet = {}
    groupby = step.groupby
    parents = [step.parentsColumn] if step.parentsColumn else []

    def _column_map(colnames: list) -> dict:
        return {col: f"${col}" for col in colnames}

    # Pipeline that will be executed to get the array of results for the starting
    # and ending milestone of the waterfall
    facet_start_end = [
        {
            "$group": {
                "_id": _column_map(groupby + [step.milestonesColumn]),
                step.valueColumn: {"$sum": f"${step.valueColumn}"},
            },
        },
        {
            "$project": {
                **{col: f"$_id.{col}" for col in groupby},
                "LABEL_waterfall": f"$_id.{step.milestonesColumn}",
                **(
                    {"GROUP_waterfall": f"$_id.{step.milestonesColumn}"}
                    if step.parentsColumn
                    else {}
                ),
                "TYPE_waterfall": None,
                step.valueColumn: 1,
                "_vqbOrder": {
                    "$cond": [{"$eq": [f"$_id.{step.milestonesColumn}", step.start]}, -1, 1]
                },
            },
        },
    ]

    # Pipeline that will be executed to get the array of results for the children
    # elements of the waterfall
    facet_children = [
        {
            "$group": {
                "_id": _column_map(
                    groupby + parents + [step.labelsColumn] + [step.milestonesColumn]
                ),
                step.valueColumn: {"$sum": f"${step.valueColumn}"},
            },
        },
        {
            "$addFields": {
                "_vqbOrder": {
                    "$cond": [{"$eq": [f"$_id.{step.milestonesColumn}", step.start]}, 1, 2]
                },
            },
        },
        {"$sort": {"_vqbOrder": 1}},
        {
            "$group": {
                "_id": {col: f"$_id.{col}" for col in groupby + parents + [step.labelsColumn]},
                "_vqbValuesArray": {"$push": f"${step.valueColumn}"},
            },
        },
        {
            "$project": {
                **{col: f"$_id.{col}" for col in groupby},
                "LABEL_waterfall": f"$_id.{step.labelsColumn}",
                **({"GROUP_waterfall": f"$_id.{step.parentsColumn}"} if step.parentsColumn else {}),
                "TYPE_waterfall": "child" if step.parentsColumn else "parent",
                step.valueColumn: {
                    "$reduce": {
                        "input": "$_vqbValuesArray",
                        "initialValue": 0,
                        "in": {"$subtract": ["$$this", "$$value"]},
                    },
                },
                "_vqbOrder": {"$literal": 0},
            },
        },
    ]

    # If.parentsColumn is define, we set the pipeline that will be executed to
    # get the array of results for the parents elements of the waterfall. In such
    # a case we add it to the concatenation of all the pipelines results arrays
    if step.parentsColumn:
        facet_parents = [
            {
                "$group": {
                    "_id": _column_map(groupby + parents + [step.milestonesColumn]),
                    step.valueColumn: {"$sum": f"${step.valueColumn}"},
                },
            },
            {
                "$addFields": {
                    "_vqbOrder": {
                        "$cond": [{"$eq": [f"$_id.{step.milestonesColumn}", step.start]}, 1, 2]
                    },
                },
            },
            {"$sort": {"_vqbOrder": 1}},
            {
                "$group": {
                    "_id": {col: f"$_id.{col}" for col in groupby + parents},
                    "_vqbValuesArray": {"$push": f"${step.valueColumn}"},
                },
            },
            {
                "$project": {
                    **{col: f"$_id.{col}" for col in groupby},
                    "LABEL_waterfall": f"$_id.{step.parentsColumn}",
                    "GROUP_waterfall": f"$_id.{step.parentsColumn}",
                    "TYPE_waterfall": "parent",
                    (step.valueColumn): {
                        "$reduce": {
                            "input": "$_vqbValuesArray",
                            "initialValue": 0,
                            "in": {"$subtract": ["$$this", "$$value"]},
                        },
                    },
                    "_vqbOrder": {"$literal": 0},
                },
            },
        ]

        facet = {
            "_vqb_start_end": facet_start_end,
            "_vqb_parents": facet_parents,
            "_vqb_children": facet_children,
        }
        concat_mongo = {"$concatArrays": ["$_vqb_start_end", "$_vqb_parents", "$_vqb_children"]}
    else:
        facet = {"_vqb_start_end": facet_start_end, "_vqb_children": facet_children}
        concat_mongo = {"$concatArrays": ["$_vqb_start_end", "$_vqb_children"]}

    return [
        {"$match": {(step.milestonesColumn): {"$in": [step.start, step.end]}}},
        {"$facet": facet},
        {"$project": {"_vqbFullArray": concat_mongo}},
        {"$unwind": "$_vqbFullArray"},
        {"$replaceRoot": {"newRoot": "$_vqbFullArray"}},
        {
            "$sort": {
                "_vqbOrder": 1,
                ("LABEL_waterfall" if step.sortBy == "label" else step.valueColumn): 1
                if step.order == "asc"
                else -1,
            },
        },
        {"$project": {"_vqbOrder": 0}},
    ]



from typing import Any

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import EvolutionStep


def translate_evolution(step: EvolutionStep) -> list[MongoStep]:
    new_column = (
        step.new_column
        if step.new_column
        else f"{step.value_col}_EVOL_{step.evolution_format.upper()}"
    )
    error_msg = "Error: More than one previous date found for the specified index columns"
    add_field_result: dict[str, Any] = {}

    if step.evolution_format == "abs":
        add_field_result[new_column] = {
            "$cond": [
                {"$eq": ["$_VQB_VALUE_PREV", "Error"]},
                error_msg,
                {"$subtract": [f"${step.value_col}", "$_VQB_VALUE_PREV"]},
            ],
        }
    else:
        add_field_result[new_column] = {
            "$switch": {
                "branches": [
                    {"case": {"$eq": ["$_VQB_VALUE_PREV", "Error"]}, "then": error_msg},
                    {"case": {"$eq": ["$_VQB_VALUE_PREV", 0]}, "then": None},
                ],
                "default": {
                    "$divide": [
                        {"$subtract": [f"${step.value_col}", "$_VQB_VALUE_PREV"]},
                        "$_VQB_VALUE_PREV",
                    ],
                },
            },
        }

    if step.evolution_type == "vsLastYear":
        unit = "year"
    elif step.evolution_type == "vsLastMonth":
        unit = "month"
    else:
        unit = "week"

    return [
        {
            "$addFields": {
                "_VQB_DATE_PREV": {
                    # NOTE: This requires mongo>=5.0
                    "$dateSubtract": {"startDate": f"${step.date_col}", "unit": unit, "amount": 1}
                }
            }
        },
        {
            "$facet": {
                "_VQB_ORIGINALS": [{"$project": {"_id": 0}}],
                "_VQB_COPIES_ARRAY": [
                    {"$group": {"_id": None, "_VQB_ALL_DOCS": {"$push": "$$ROOT"}}}
                ],
            },
        },
        {"$unwind": "$_VQB_ORIGINALS"},
        {
            "$project": {
                "_VQB_ORIGINALS": {
                    "$mergeObjects": [
                        "$_VQB_ORIGINALS",
                        {"$arrayElemAt": ["$_VQB_COPIES_ARRAY", 0]},
                    ],
                },
            },
        },
        {"$replaceRoot": {"newRoot": "$_VQB_ORIGINALS"}},
        {
            "$addFields": {
                "_VQB_ALL_DOCS": {
                    "$filter": {
                        "input": "$_VQB_ALL_DOCS",
                        "as": "item",
                        "cond": {
                            "$and": [
                                {"$eq": ["$_VQB_DATE_PREV", f"$$item.{step.date_col}"]},
                                {"$eq": [f"${col}", f"$$item.{col}"] for col in step.index_columns},
                            ],
                        },
                    },
                },
            },
        },
        {
            "$addFields": {
                "_VQB_VALUE_PREV": {
                    "$cond": [
                        {"$gt": [{"$size": f"$_VQB_ALL_DOCS.{step.value_col}"}, 1]},
                        "Error",
                        {"$arrayElemAt": [f"$_VQB_ALL_DOCS.{step.value_col}", 0]},
                    ],
                },
            },
        },
        {"$addFields": add_field_result},
        {
            "$project": {
                "_VQB_ALL_DOCS": 0,
                "_VQB_DATE_PREV": 0,
                "_VQB_VALUE_PREV": 0,
            },
        },
    ]



from typing import TYPE_CHECKING, Literal, Union

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.combination import (
    Reference,
    ReferenceResolver,
    resolve_if_reference,
)

if TYPE_CHECKING:
    from weaverbird.pipeline.pipeline import PipelineWithVariables


class BaseDomainStep(BaseStep):
    name: Literal["domain"] = "domain"


class DomainStep(BaseDomainStep):
    domain: str


class DomainStepWithRef(BaseDomainStep):
    domain: str | Reference

    async def resolve_references(
        self, reference_resolver: ReferenceResolver
    ) -> Union[DomainStep, "PipelineWithVariables"]:
        """
        This resolution can return a whole pipeline, which needs to replace the step.
        Not that the resulting array must be flattened:
        it should look like [step 1, step 2, step 3], not [[step 1, step 2], step 3]
        """
        from weaverbird.pipeline.references import PipelineWithRefs

        resolved = await resolve_if_reference(reference_resolver, self.domain)
        if isinstance(resolved, list):
            return await PipelineWithRefs(steps=resolved).resolve_references(reference_resolver)
        elif resolved is None:
            from weaverbird.pipeline.references import ReferenceUnresolved

            raise ReferenceUnresolved()
        else:
            return DomainStep(
                name=self.name,
                domain=resolved,
            )



from typing import Any, Literal, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseModel, BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName

from ..conditions import Condition


class IfThenElse(BaseModel):
    condition: Condition = Field(alias="if")
    then: Any
    # We can't use | with a quoted type in python 3.11
    else_value: Union["IfThenElse", Any] = Field(alias="else")
    # NOTE: Some existing pipelines may already have defined this in a nested ifthenelse block. Our
    # BaseStep models forbids extra values, so we add this here too
    name: Literal["ifthenelse"] = "ifthenelse"


IfThenElse.update_forward_refs()


class IfthenelseStep(BaseStep, IfThenElse):
    name: Literal["ifthenelse"] = "ifthenelse"
    new_column: ColumnName


class IfThenElseStepWithVariables(IfthenelseStep, StepWithVariablesMixin):
    ...



from pypika import functions
from pypika.dialects import PostgreSQLQuery
from pypika.queries import Selectable

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateFormatMapping,
    SQLTranslator,
)


class PostgreSQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.POSTGRES
    QUERY_CLS = PostgreSQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://www.postgresql.org/docs/current/functions-formatting.html
        day_number="DD",
        month_number="MM",
        month_short="Mon",
        # FM prefix prevents postgres from padding the month to 9 chars
        month_full="FMMonth",
        year="YYYY",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.SIMILAR_TO
    TO_DATE_OP = ToDateOp.TO_TIMESTAMP

    @classmethod
    def _interval_to_seconds(cls, value: Selectable) -> functions.Function:
        return functions.Extract("epoch", value)


SQLTranslator.register(PostgreSQLTranslator)



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.top import TopStep


def translate_top(step: TopStep) -> list[MongoStep]:
    sort_order = 1 if step.sort == "asc" else -1
    group_cols = {g: f"${g}" for g in step.groups} if step.groups else None
    return [
        {"$sort": {step.rank_on: sort_order}},
        {"$group": {"_id": group_cols, "_vqbAppArray": {"$push": "$$ROOT"}}},
        {"$project": {"_vqbAppTopElems": {"$slice": ["$_vqbAppArray", step.limit]}}},
        {"$unwind": "$_vqbAppTopElems"},
        {"$replaceRoot": {"newRoot": "$_vqbAppTopElems"}},
    ]



from datetime import datetime
from typing import TYPE_CHECKING, TypeVar

from pypika import Field, Table, functions
from pypika.dialects import MySQLQuery
from pypika.enums import Dialects
from pypika.terms import Case, Term
from pypika.utils import format_quotes

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateFormatMapping,
    SQLTranslator,
    StepContext,
)
from weaverbird.pipeline.steps import UnpivotStep

Self = TypeVar("Self", bound="MySQLTranslator")

if TYPE_CHECKING:
    from pypika.queries import QueryBuilder

    from weaverbird.pipeline.steps import DateExtractStep, SplitStep


class MySQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.MYSQL
    QUERY_CLS = MySQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="SIGNED",
        date="DATE",
        float="DOUBLE",
        integer="UNSIGNED",
        text="CHAR",
        datetime="DATETIME",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://dev.mysql.com/doc/refman/8.0/en/date-and-time-functions.html#function_date-format
        day_number="%d",
        month_number="%m",
        month_short="%b",
        month_full="%M",
        year="%Y",
    )
    # Requires MySQL>=8
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = False
    FROM_DATE_OP = FromDateOp.DATE_FORMAT
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.STR_TO_DATE

    @classmethod
    def _build_unpivot_col(
        cls, *, step: "UnpivotStep", quote_char: str | None, secondary_quote_char: str
    ) -> str:
        value_col = format_quotes(step.value_column_name, quote_char)
        unpivot_col = format_quotes(step.unpivot_column_name, quote_char)
        in_cols = [format_quotes(col, quote_char) for col in step.unpivot]
        in_single_quote_cols = [format_quotes(col, secondary_quote_char) for col in step.unpivot]
        rows = " UNION ALL SELECT ".join(
            [f"{col}, {val}" for col, val in zip(in_cols, in_single_quote_cols, strict=True)]
        )
        return f" t1 CROSS JOIN LATERAL(SELECT {rows}) AS t2({value_col}, {unpivot_col})"

    def split(
        self: Self,
        *,
        builder: "QueryBuilder",
        prev_step_table: str,
        columns: list[str],
        step: "SplitStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_table)[step.column]
        new_cols = [f"{step.column}_{i + 1}" for i in range(step.number_cols_to_keep)]

        def build_columns():
            for i in range(step.number_cols_to_keep):
                yield (
                    # We need a case, because by default, the nested SubstringIndex will return the
                    # entire string: -1 (the last part of the split) will be index 0. So we first
                    # check if the delimiter is even contained in the regex
                    Case()
                    .when(
                        # This is hacky but I couldn't figure another way. This regex ensures that
                        # the group matches at least i times (i.e., the delimiter is present at
                        # least n times). The regex is always true for i=0 so kind of unnecessary,
                        # but having another Case statement for i=0 would be hell
                        col_field.regexp(f"(({step.delimiter}).*){{{i}}}"),
                        # https://stackoverflow.com/a/32500349
                        SubstringIndex(
                            SubstringIndex(col_field, step.delimiter, i + 1), step.delimiter, -1
                        ),
                    )
                    .else_("")
                    .as_(new_cols[i])
                )

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_table).select(
            *columns, *build_columns()
        )
        return StepContext(query, columns + new_cols)

    @staticmethod
    def _cast_to_timestamp(value: str | datetime | Field | Term) -> functions.Function:
        return functions.Timestamp(value)

    @classmethod
    def _add_date(
        cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None
    ) -> Term:
        return super()._add_date(
            target_column=target_column,
            duration=duration,
            unit=unit,
            dialect=dialect or Dialects.MYSQL,
        )

    def dateextract(
        self: Self,
        *,
        builder: "QueryBuilder",
        prev_step_table: str,
        columns: list[str],
        step: "DateExtractStep",
    ) -> StepContext:
        raise NotImplementedError


SQLTranslator.register(MySQLTranslator)


class SubstringIndex(functions.Function):
    def __init__(
        self, term: str | Field, delimiter: str, count: int, alias: str | None = None
    ) -> None:
        super().__init__("SUBSTRING_INDEX", term, delimiter, count, alias=alias)



from typing import Any, Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

TYPE_WATERFALL_COLUMN = "TYPE_waterfall"
LABEL_WATERFALL_COLUMN = "LABEL_waterfall"
GROUP_WATERFALL_COLUMN = "GROUP_waterfall"

_RESULT_COLUMN = "result"


class WaterfallStep(BaseStep):
    name: Literal["waterfall"] = "waterfall"
    valueColumn: ColumnName
    milestonesColumn: ColumnName
    start: Any
    end: Any
    labelsColumn: ColumnName
    sortBy: Literal["value", "label"]
    order: Literal["desc", "asc"]
    parentsColumn: ColumnName | None
    groupby: list[ColumnName] = []


class WaterfallStepWithVariable(WaterfallStep, StepWithVariablesMixin):
    groupby: TemplatedVariable | list[TemplatedVariable]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import RollupStep

_ID_COLUMN = "_id"
_NEW_ID_COLUMN = "__id"


def column_map(s: list[str]) -> dict[str, str]:
    return {e: f"${e}" for e in s}


def translate_rollup(step: RollupStep) -> list[MongoStep]:
    facet: dict[str, list[MongoStep]] = {}
    label_col = step.label_col or "label"
    level_col = step.level_col or "level"
    parent_label_col = step.parent_label_col or "parent"
    add_fields = {}

    for idx, elem in enumerate(step.hierarchy):
        id = column_map(step.hierarchy[: idx + 1] + (step.groupby or []))
        aggs: dict[str, dict] = {}
        for agg_step in step.aggregations:
            cols = agg_step.columns
            # _id is a name reserved by mongo. If for some reason, a user wants to aggregate the _id
            # column, the aggregation result will be stored in a new __id column
            new_cols = [
                col if col != _ID_COLUMN else _NEW_ID_COLUMN for col in agg_step.new_columns
            ]

            if agg_step.agg_function == "count":
                for i in range(len(cols)):
                    # cols and new_cols are always of same length
                    aggs[new_cols[i]] = {"$sum": 1}
            elif agg_step.agg_function == "count distinct":
                # specific step needed to count distinct values
                for i in range(len(cols)):
                    # build a set of unique values
                    aggs[new_cols[i]] = {"$addToSet": f"${cols[i]}"}
                    # count the number of items in the set
                    add_fields[new_cols[i]] = {"$size": f"${new_cols[i]}"}
            else:
                for i in range(len(cols)):
                    # cols and new_cols are always of same length
                    aggs[new_cols[i]] = {f"${agg_step.agg_function}": f"${cols[i]}"}

        project: dict = {
            _ID_COLUMN: 0,
            **{k: f"${_ID_COLUMN}.{k}" for k in id.keys()},
            **{k: 1 for k in aggs.keys()},
            label_col: f"${_ID_COLUMN}.{elem}",
            level_col: elem,
        }

        if idx > 0:
            project[parent_label_col] = f"${_ID_COLUMN}.{step.hierarchy[idx - 1]}"

        add_fields_to_add_to_pipeline = [{"$addFields": add_fields}] if add_fields else []

        facet[f"level_{idx}"] = [
            {
                "$group": {
                    _ID_COLUMN: id,
                    **aggs,
                },
            },
            *add_fields_to_add_to_pipeline,
            {
                "$project": project,
            },
        ]

    return [
        {"$facet": facet},
        {
            "$project": {
                "_vqbRollupLevels": {
                    "$concatArrays": [f"${k}" for k in sorted(facet.keys())],
                },
            },
        },
        {"$unwind": "$_vqbRollupLevels"},
        {"$replaceRoot": {"newRoot": "$_vqbRollupLevels"}},
    ]



from collections.abc import Iterable
from typing import Annotated, Any

from pydantic import BaseModel, Field

from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
)
from weaverbird.pipeline.steps.hierarchy import HierarchyStep

from .steps import (
    AbsoluteValueStep,
    AbsoluteValueStepWithVariable,
    AddMissingDatesStep,
    AddMissingDatesStepWithVariables,
    AggregateStep,
    AggregateStepWithVariables,
    AppendStep,
    AppendStepWithVariable,
    ArgmaxStep,
    ArgmaxStepWithVariable,
    ArgminStep,
    ArgminStepWithVariable,
    CompareTextStep,
    CompareTextStepWithVariables,
    ConcatenateStep,
    ConcatenateStepWithVariable,
    ConvertStep,
    CumSumStep,
    CumSumStepWithVariable,
    CustomSqlStep,
    CustomStep,
    DateExtractStep,
    DateExtractStepWithVariable,
    DeleteStep,
    DissolveStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    DurationStepWithVariable,
    EvolutionStep,
    EvolutionStepWithVariable,
    FillnaStep,
    FillnaStepWithVariable,
    FilterStep,
    FilterStepWithVariables,
    FormulaStep,
    FormulaStepWithVariable,
    FromdateStep,
    IfthenelseStep,
    IfThenElseStepWithVariables,
    JoinStep,
    JoinStepWithVariable,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    PivotStepWithVariable,
    RankStep,
    RankStepWithVariable,
    RenameStep,
    RenameStepWithVariable,
    ReplaceStep,
    ReplaceStepWithVariable,
    ReplaceTextStep,
    ReplaceTextStepWithVariable,
    RollupStep,
    RollupStepWithVariable,
    SelectStep,
    SimplifyStep,
    SortStep,
    SplitStep,
    SplitStepWithVariable,
    StatisticsStep,
    SubstringStep,
    TextStep,
    TextStepWithVariable,
    ToDateStep,
    TopStep,
    TopStepWithVariables,
    TotalsStep,
    TotalsStepWithVariable,
    TrimStep,
    UniqueGroupsStep,
    UniqueGroupsStepWithVariable,
    UnpivotStep,
    UnpivotStepWithVariable,
    UppercaseStep,
    WaterfallStep,
    WaterfallStepWithVariable,
)

PipelineStep = Annotated[
    AbsoluteValueStep
    | AddMissingDatesStep
    | AggregateStep
    | AppendStep
    | ArgmaxStep
    | ArgminStep
    | CompareTextStep
    | ConcatenateStep
    | ConvertStep
    | CumSumStep
    | CustomSqlStep
    | CustomStep
    | DateExtractStep
    | DeleteStep
    | DissolveStep
    | DomainStep
    | DuplicateStep
    | DurationStep
    | EvolutionStep
    | FillnaStep
    | FilterStep
    | FormulaStep
    | FromdateStep
    | FromdateStep
    | HierarchyStep
    | IfthenelseStep
    | JoinStep
    | LowercaseStep
    | MovingAverageStep
    | PercentageStep
    | PivotStep
    | RankStep
    | RenameStep
    | ReplaceStep
    | ReplaceTextStep
    | RollupStep
    | SelectStep
    | SimplifyStep
    | SortStep
    | SplitStep
    | StatisticsStep
    | SubstringStep
    | TextStep
    | ToDateStep
    | TopStep
    | TotalsStep
    | TrimStep
    | UniqueGroupsStep
    | UnpivotStep
    | UppercaseStep
    | WaterfallStep,
    Field(discriminator="name"),  # noqa: F821
]


class Pipeline(BaseModel):
    steps: list[PipelineStep]

    def dict(self, *, exclude_none: bool = True, **kwargs) -> dict:
        return super().dict(exclude_none=True, **kwargs)


PipelineStepWithVariables = Annotated[
    AbsoluteValueStepWithVariable
    | AddMissingDatesStepWithVariables
    | AggregateStepWithVariables
    | AppendStepWithVariable
    | ArgmaxStepWithVariable
    | ArgminStepWithVariable
    | CompareTextStepWithVariables
    | ConcatenateStepWithVariable
    | CumSumStepWithVariable
    | DateExtractStepWithVariable
    | DurationStepWithVariable
    | EvolutionStepWithVariable
    | FillnaStepWithVariable
    | FilterStepWithVariables
    | FormulaStepWithVariable
    | IfThenElseStepWithVariables
    | JoinStepWithVariable
    | PivotStepWithVariable
    | RankStepWithVariable
    | RenameStepWithVariable
    | ReplaceStepWithVariable
    | ReplaceTextStepWithVariable
    | RollupStepWithVariable
    | SplitStepWithVariable
    | TextStepWithVariable
    | TopStepWithVariables
    | TotalsStepWithVariable
    | UniqueGroupsStepWithVariable
    | UnpivotStepWithVariable
    | WaterfallStepWithVariable,
    Field(discriminator="name"),  # noqa: F821
]

VOID_REPR = "__VOID__"


def _remove_void_from_combo_condition(
    condition: ConditionComboAnd | ConditionComboOr,
) -> ConditionComboAnd | ConditionComboOr | None:
    """
    Loop into a combo condition and search for embedded condition that may
    contains __VOID__ as column/value
    """
    sanitized_conditions: list[Condition] = []
    combo = "and_" if hasattr(condition, "and_") else "or_"

    for cond in getattr(condition, combo) or []:
        if (transformed_condition := _remove_void_from_condition(cond)) is not None:
            sanitized_conditions.append(transformed_condition)
    if len(sanitized_conditions) == 0:
        return None

    setattr(condition, combo, sanitized_conditions)

    return condition


def _remove_void_from_condition(condition: Condition) -> Condition | None:
    """
    For a given condition either it's a combo or a simple condition
    recursively check for it's columns and values and remove the one
    with column/value = __VOID__

    """
    if isinstance(condition, ConditionComboAnd | ConditionComboOr):
        condition = _remove_void_from_combo_condition(condition)
    elif isinstance(condition, ComparisonCondition | MatchCondition | DateBoundCondition):
        if condition.column == VOID_REPR or condition.value == VOID_REPR:
            return None
    elif isinstance(condition, InclusionCondition):
        condition_values = [v for v in condition.value if v != VOID_REPR]
        if len(condition_values) == 0:
            return None
        condition.value = condition_values

    return condition


def remove_void_conditions_from_filter_steps(
    steps: list[PipelineStepWithVariables | PipelineStep],
) -> list[PipelineStepWithVariables | PipelineStep]:
    """
    This method will remove all FilterStep with conditions having "__VOID__"
    in them. either the "value" key or the "column" key.
    """

    final_steps = []
    for step in steps:
        if isinstance(step, FilterStep):
            if (condition := _remove_void_from_condition(step.condition)) is not None:
                final_steps.append(FilterStep(condition=condition))
        else:
            final_steps.append(step)

    return final_steps


def _is_empty(data: Any) -> bool:
    if isinstance(data, list | dict):
        return not bool(data)
    return False


def _remove_empty_elements_from_match_steps(data: Any, depth: int = 0) -> tuple[Any, bool]:
    """Returns the passed data with empty elements removed from $match steps.

    The passed data is returned along with a boolean indicating wether it is empty
    """
    if isinstance(data, dict):
        data_transformed: dict[str, Any] | list[Any] = {}

        for k, v in data.items():
            # If we are at the root level and the step is not a match, skip it
            if depth == 0 and k != "$match":
                data_transformed[k] = v
                continue
            cleaned, is_empty = _remove_empty_elements_from_match_steps(v, depth=depth + 1)
            if not is_empty:
                data_transformed[k] = cleaned

        return data_transformed, _is_empty(data_transformed)

    elif isinstance(data, list):
        # NOTE: Some of our steps, such as rank or addmissingdates, set some values in the pipeline
        # to an empty list. In consequence, a sanitized list should only be considered empty if it
        # was not already empty before the transformation
        if len(data) < 1:
            return [], False
        data_transformed = [
            elem
            for (elem, is_empty) in (
                # Only increasing depth for dicts
                _remove_empty_elements_from_match_steps(item, depth=depth)
                for item in data
            )
            if not is_empty
        ]

        return data_transformed, _is_empty(data_transformed)
    else:
        return data, _is_empty(data)


def _remove_void_entries_from_dict(d: dict[str, Any]) -> Iterable[tuple[str, Any]]:
    for key, val in d.items():
        if isinstance(val, str):
            if val.strip() == VOID_REPR:
                continue
            yield key, val
        else:
            yield key, _remove_void_entries(val)


def _remove_void_entries(
    mongo_steps: dict[str, Any] | list | None,
) -> dict[str, Any] | list | None:
    """
    This method will remove element with value string as "__VOID__"
    """

    if isinstance(mongo_steps, dict):
        return dict(_remove_void_entries_from_dict(mongo_steps))
    elif isinstance(mongo_steps, list):
        return [_remove_void_entries(step) for step in mongo_steps]
    else:
        return mongo_steps


def _is_empty_match_column(elem: Any):
    # Not matching None here, because we don't want to consider {'$match': {'col': None}} to be
    # empty
    if elem == {}:
        return True
    if isinstance(elem, dict):
        return _is_match_empty(elem)

    return False


def _is_match_empty(match_: dict) -> bool:
    return all(_is_empty_match_column(v) for v in match_.values())


def _is_match_statement(d: Any) -> bool:
    return isinstance(d, dict) and list(d.keys()) == ["$match"]


def _sanitize_match(query: dict) -> dict:
    if _is_match_empty(query):
        return {}
    if "$and" in query:
        and_condition = query["$and"]
        if isinstance(and_condition, list):
            query["$and"] = [elem for elem in and_condition if not _is_empty_match_column(elem)]
    return query


def _sanitize_query_matches(query: dict | list[dict]) -> Any:
    """Transforms match operations matching nothing into match-alls.

    If a $match would match nothing (for example, {'$match': {'field': {}}}), transform into a
    passthrough. It cannot be removed from the query to prevent having an empty query.
    """
    if isinstance(query, list):
        # we need to have $match as first step here
        if bool(query) and "$match" not in query[0]:
            query = [{"$match": {}}] + query

        return [
            {"$match": _sanitize_match(q["$match"])} if _is_match_statement(q) else q for q in query
        ]
    return query


def remove_void_conditions_from_mongo_steps(
    mongo_steps: dict | list[dict],
) -> dict | list[dict]:
    without_voids = _remove_void_entries(mongo_steps) or []
    without_empty_elements, _ = _remove_empty_elements_from_match_steps(without_voids)
    return _sanitize_query_matches(without_empty_elements)


# TODO move to a dedicated variables module
class PipelineWithVariables(BaseModel):
    steps: list[PipelineStepWithVariables | PipelineStep]

    def render(self, variables: dict[str, Any], renderer) -> Pipeline:
        # TODO it must be more efficient to render the full pipeline once
        steps_rendered = [
            step.render(variables, renderer) if hasattr(step, "render") else step  # type: ignore
            for step in self.steps
        ]
        return Pipeline(steps=steps_rendered)


PipelineWithVariables.update_forward_refs()



from datetime import datetime
from typing import Literal

from pydantic import BaseModel


class RelativeDate(BaseModel):
    date: datetime | str | None
    operator: Literal["from", "until", "before", "after"]
    quantity: int
    duration: Literal["year", "quarter", "month", "week", "day"]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ConvertStep

TYPE_MAP = {
    "boolean": "bool",
    "date": "date",
    "float": "double",
    "integer": "long",  # better to cast into 64-bit integers. 32-bit integers ('int') cannot be casted to dates
    "text": "string",
}


def translate_convert(step: ConvertStep) -> list[MongoStep]:
    # /!\ we want to get the same results than with the mongo typescript
    # translator, not than with the pandas executor.

    mongo_add_fields = {}
    mongo_type = TYPE_MAP[step.data_type] if step.data_type in TYPE_MAP else ""
    for column in step.columns:
        # Mongo cannot cast integers into date but only long into date, so we
        # manage the cast from int to long when needed.
        _input: dict | str
        if mongo_type == "date":
            _input = {
                "$cond": [
                    {"$eq": [{"$type": f"${column}"}, "int"]},
                    {"$toLong": f"${column}"},
                    f"${column}",
                ]
            }
        else:
            _input = f"${column}"
        mongo_add_fields[column] = {
            "$convert": {
                "input": _input,
                "to": mongo_type,
                "onError": None,  # failed conversions result in null instead of crashing
            }
        }

    return [{"$addFields": mongo_add_fields}]



from collections.abc import Awaitable, Callable
from typing import Literal

from pydantic import BaseModel


class Reference(BaseModel):
    type: Literal["ref"] = "ref"
    uid: str

    # Adding a prefix so is does not have the same hash as a domain whose name would be the same as
    # the uid
    def __hash__(self):
        return hash(f"__ref__{self.uid}")


PipelineOrDomainName = list[dict] | str  # can be either a domain name or a complete pipeline
PipelineOrDomainNameOrReference = PipelineOrDomainName | Reference

# A reference returning None means that it should be skipped
ReferenceResolver = Callable[[Reference], Awaitable[PipelineOrDomainName | None]]


async def resolve_if_reference(
    reference_resolver: ReferenceResolver,
    pipeline_or_domain_name_or_ref: PipelineOrDomainNameOrReference,
) -> PipelineOrDomainName | None:
    from weaverbird.pipeline.references import PipelineWithRefs

    if isinstance(pipeline_or_domain_name_or_ref, Reference):
        pipeline_or_domain_name = await reference_resolver(pipeline_or_domain_name_or_ref)
        if isinstance(pipeline_or_domain_name, list):
            # Recursively resolve any reference in sub-pipelines
            pipeline = PipelineWithRefs(steps=pipeline_or_domain_name)
            from weaverbird.pipeline.references import ReferenceUnresolved

            try:
                pipeline_without_refs = await pipeline.resolve_references(reference_resolver)
                return pipeline_without_refs.dict()["steps"]
            except ReferenceUnresolved:
                return None  # skip
        else:
            return pipeline_or_domain_name
    else:
        return pipeline_or_domain_name_or_ref



import re
from collections import defaultdict

from pandas import DataFrame

from weaverbird.pipeline.types import ColumnName

REGEX_COLUMN_NAME_WITH_SUFFIX = re.compile(r"(.*)_(\d+)")


def rename_duplicated_columns(df: DataFrame) -> DataFrame:
    """
    If a dataframe has several columns with the same name,
    adds a suffix to the duplicates (_1, _2, etc.)
    (the first occurrence remains without suffix)
    If there are already some columns suffixed with numbers
    in the dataframe, then the new suffix numbers will start
    after the current maximum.
    E.g: columns ['foo', 'foo_2', 'foo'] -> ['foo', 'foo_2', 'foo_3']

    Processes inplace.
    """
    old_cols = list(df.columns)
    max_suffix: defaultdict[ColumnName, int] = defaultdict(lambda: 0)

    # First step: find the current max suffix for each column name
    for column in old_cols:
        if match := REGEX_COLUMN_NAME_WITH_SUFFIX.fullmatch(column):
            column_basename, suffix = match.groups()
            suffix = int(suffix)
        else:
            column_basename, suffix = column, 0
        max_suffix[column_basename] = max(max_suffix[column_basename], suffix + 1)

    new_cols = []

    def _find_new_name_for_column(column: str) -> str:
        if column not in new_cols:
            return column

        if match := REGEX_COLUMN_NAME_WITH_SUFFIX.fullmatch(column):
            column_basename, suffix = match.groups()
            new_suffix = max_suffix[column_basename]
        else:
            column_basename, new_suffix = column, max_suffix[column]
        new_column = f"{column_basename}_{new_suffix}"
        max_suffix[column_basename] += 1  # increment the current max suffix
        return new_column

    # Second step: use this max suffix to rename the columns where needed:
    for column in old_cols:
        new_col = _find_new_name_for_column(column)
        new_cols.append(new_col)

    # Edit the column names (inplace):
    df.columns = new_cols
    return df



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DomainStep


def translate_domain(step: DomainStep) -> list[MongoStep]:
    return [{"$match": {"domain": step.domain}}]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DurationStep

# A mapping of multiplier to apply to convert milliseconds in days, hours, minutes or seconds
DURATION_MULTIPLIER_MAP = {
    "days": 24 * 60 * 60 * 1000,
    "hours": 60 * 60 * 1000,
    "minutes": 60 * 1000,
    "seconds": 1000,
}


def translate_duration(step: DurationStep) -> list[MongoStep]:
    return [
        {
            "$addFields": {
                step.new_column_name: {
                    "$divide": [
                        # a time difference between dates is returned in milliseconds by Mongo
                        {"$subtract": [f"${step.end_date_column}", f"${step.start_date_column}"]},
                        DURATION_MULTIPLIER_MAP[step.duration_in],
                    ],
                },
            },
        }
    ]



from pandas import DataFrame, merge

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import JoinStep

from .utils.cleaning import rename_duplicated_columns
from .utils.combination import resolve_pipeline_for_combination


def execute_join(
    step: JoinStep,
    df: DataFrame,
    domain_retriever: DomainRetriever,
    execute_pipeline: PipelineExecutor,
) -> DataFrame:
    right_df = resolve_pipeline_for_combination(
        step.right_pipeline, domain_retriever, execute_pipeline
    )

    if step.type == "left outer":
        how = "outer"
    else:
        how = step.type

    result = merge(
        df,
        right_df,
        left_on=[o[0] for o in step.on],
        right_on=[o[1] for o in step.on],
        how=how,
        suffixes=("", "_JOIN"),
    )
    return rename_duplicated_columns(result)



import numpy as np
from pandas import DataFrame

from weaverbird.backends.pandas_executor.steps.utils.formula import eval_formula
from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.ifthenelse import IfThenElse, IfthenelseStep

from .utils.condition import apply_condition


def _execute_ifthenelse(ifthenelse: IfThenElse, df: DataFrame, new_column) -> DataFrame:
    if isinstance(ifthenelse.else_value, IfThenElse):
        else_branch = _execute_ifthenelse(ifthenelse.else_value, df, new_column)[new_column]
    else:
        else_branch = eval_formula(df, str(ifthenelse.else_value))

    then_branch = eval_formula(df, str(ifthenelse.then))
    return df.assign(
        **{
            new_column: np.where(
                apply_condition(ifthenelse.condition, df), then_branch, else_branch
            )
        }
    )


def execute_ifthenelse(
    step: IfthenelseStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    return _execute_ifthenelse(step, df, step.new_column)



from typing import Literal

from pydantic import validator

from weaverbird.pipeline.steps.utils.base import BaseStep


class CustomSqlStep(BaseStep):
    name: Literal["customsql"] = "customsql"
    query: str

    @staticmethod
    def _strip_query(query: str) -> str:
        return query.strip().strip(";")

    @validator("query")
    def _validate_query(cls, query: str) -> str:
        assert ";" not in (
            stripped := cls._strip_query(query)
        ), "Custom SQL queries must not contain semicolumns"
        return stripped


# /!\ Do not create CustomSqlStepWithVariables
# (variables should not be rendered using nosql_apply_parameters_to_query)



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.steps import FilterStep

operator_mapping = {
    "eq": "$eq",
    "ne": "$ne",
    "lt": "$lt",
    "le": "$lte",
    "gt": "$gt",
    "ge": "$gte",
    "in": "$in",
    "nin": "$nin",
    "isnull": "$eq",
    "notnull": "$ne",
    "from": "$gte",
    "until": "$lte",
}

RELATIVE_DATE_OPERATORS: dict[str, int] = {"until": -1, "from": +1}


def translate_relative_date(value: RelativeDate):
    sign = RELATIVE_DATE_OPERATORS[value.operator]
    return {
        "$dateAdd": {
            "startDate": truncate_to_day({"$toDate": value.date}),
            "amount": sign * abs(int(value.quantity)),
            "unit": value.duration,
        },
    }


def truncate_to_day(value):
    return {
        "$dateTrunc": {
            "date": value,
            "unit": "day",
        }
    }


def build_match_tree(condition: Condition, parent_operator="and") -> dict:
    # coumpound conditions

    if isinstance(condition, ConditionComboAnd):
        return {"$and": [build_match_tree(c, "and") for c in condition.and_]}
    elif isinstance(condition, ConditionComboOr):
        return {"$or": [build_match_tree(c, "or") for c in condition.or_]}

    # simple conditions
    elif isinstance(condition, ComparisonCondition) or isinstance(condition, InclusionCondition):
        return {condition.column: {operator_mapping[condition.operator]: condition.value}}
    elif isinstance(condition, NullCondition):
        return {condition.column: {operator_mapping[condition.operator]: None}}

    elif isinstance(condition, MatchCondition):
        if condition.operator == "matches":
            return {condition.column: {"$regex": condition.value, "$options": "i"}}
        elif condition.operator == "notmatches":
            return {condition.column: {"$not": {"$regex": condition.value, "$options": "i"}}}

    # dates
    elif isinstance(condition, DateBoundCondition):
        if isinstance(condition.value, RelativeDate):
            target_date = translate_relative_date(condition.value)
        else:
            target_date = {"$toDate": condition.value}

        return {
            "$expr": {
                operator_mapping[condition.operator]: [
                    truncate_to_day(f"${condition.column}"),
                    truncate_to_day(target_date),
                ]
            }
        }

    # mypy
    raise NotImplementedError(f"build_match_tree not implemented for condition: {condition}")


def translate_filter(step: FilterStep) -> list[MongoStep]:
    return [{"$match": build_match_tree(step.condition)}]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.fillna import FillnaStep


def translate_fillna(step: FillnaStep) -> list[MongoStep]:
    add_fields = {col: {"$ifNull": [f"${col}", step.value]} for col in step.columns}
    return [{"$addFields": add_fields}]



from abc import ABC
from datetime import datetime
from typing import Annotated, Any, Literal

from pydantic import BaseConfig, BaseModel, Field

from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.types import ColumnName


class BaseCondition(BaseModel):
    ...


class ComparisonCondition(BaseCondition):
    column: ColumnName
    operator: Literal["eq", "ne", "lt", "le", "gt", "ge"]
    value: Any


class InclusionCondition(BaseCondition):
    column: ColumnName
    operator: Literal["in", "nin"]
    value: list[Any]


class NullCondition(BaseCondition):
    column: ColumnName
    operator: Literal["isnull", "notnull"]


class MatchCondition(BaseCondition):
    column: ColumnName
    operator: Literal["matches", "notmatches"]
    value: str


class DateBoundCondition(BaseModel):
    column: ColumnName
    operator: Literal["from", "until"]
    value: RelativeDate | datetime | str


SimpleCondition = Annotated[
    ComparisonCondition | InclusionCondition | NullCondition | MatchCondition | DateBoundCondition,
    Field(discriminator="operator"),  # noqa: F821
]


class BaseConditionCombo(BaseCondition, ABC):
    class Config(BaseConfig):
        allow_population_by_field_name = True

    def to_dict(self):
        return self.dict(by_alias=True)


class ConditionComboAnd(BaseConditionCombo):
    and_: list["Condition"] = Field(..., alias="and")


class ConditionComboOr(BaseConditionCombo):
    or_: list["Condition"] = Field(..., alias="or")


Condition = ConditionComboAnd | ConditionComboOr | SimpleCondition
ConditionComboOr.update_forward_refs()
ConditionComboAnd.update_forward_refs()



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import TemplatedVariable

EVOLUTION_TYPE = Literal["vsLastYear", "vsLastMonth", "vsLastWeek", "vsLastDay"]
EVOLUTION_FORMAT = Literal["abs", "pct"]


class EvolutionStep(BaseStep):
    name: Literal["evolution"] = "evolution"
    date_col: str
    value_col: str
    evolution_type: EVOLUTION_TYPE
    evolution_format: EVOLUTION_FORMAT
    index_columns: list[str] = Field(default_factory=list)
    new_column: str | None


class EvolutionStepWithVariable(EvolutionStep, StepWithVariablesMixin):
    index_columns: TemplatedVariable | list[TemplatedVariable] = Field(default_factory=list)



from weaverbird.backends.pandas_executor.steps.replacetext import execute_replacetext

from ..types import StepExecutor
from .absolutevalue import execute_absolutevalue
from .addmissingdates import execute_addmissingdates
from .aggregate import execute_aggregate
from .append import execute_append
from .argmax import execute_argmax
from .argmin import execute_argmin
from .comparetext import execute_comparetext
from .concatenate import execute_concatenate
from .convert import execute_convert
from .cumsum import execute_cumsum
from .date_extract import execute_date_extract
from .delete import execute_delete
from .dissolve import execute_dissolve
from .domain import execute_domain
from .duplicate import execute_duplicate
from .duration import execute_duration
from .evolution import execute_evolution
from .fillna import execute_fillna
from .filter import execute_filter
from .formula import execute_formula
from .fromdate import execute_fromdate
from .hierarchy import execute_hierarchy
from .ifthenelse import execute_ifthenelse
from .join import execute_join
from .lowercase import execute_lowercase
from .moving_average import execute_moving_average
from .percentage import execute_percentage
from .pivot import execute_pivot
from .rank import execute_rank
from .rename import execute_rename
from .replace import execute_replace
from .rollup import execute_rollup
from .select import execute_select
from .simplify import execute_simplify
from .sort import execute_sort
from .split import execute_split
from .statistics import execute_statistics
from .substring import execute_substring
from .text import execute_text
from .todate import execute_todate
from .top import execute_top
from .totals import execute_totals
from .trim import execute_trim
from .uniquegroups import execute_uniquegroups
from .unpivot import execute_unpivot
from .uppercase import execute_uppercase
from .waterfall import execute_waterfall

steps_executors: dict[str, StepExecutor] = {
    "absolutevalue": execute_absolutevalue,
    "addmissingdates": execute_addmissingdates,
    "aggregate": execute_aggregate,
    "append": execute_append,  # type: ignore
    "argmax": execute_argmax,
    "argmin": execute_argmin,
    "comparetext": execute_comparetext,
    "concatenate": execute_concatenate,
    "convert": execute_convert,
    "cumsum": execute_cumsum,
    "dateextract": execute_date_extract,
    "delete": execute_delete,
    "dissolve": execute_dissolve,
    "domain": execute_domain,  # type: ignore
    "duplicate": execute_duplicate,
    "duration": execute_duration,
    "evolution": execute_evolution,
    "fillna": execute_fillna,
    "filter": execute_filter,
    "formula": execute_formula,
    "fromdate": execute_fromdate,
    "hierarchy": execute_hierarchy,
    "ifthenelse": execute_ifthenelse,
    "join": execute_join,  # type: ignore
    "lowercase": execute_lowercase,
    "movingaverage": execute_moving_average,
    "percentage": execute_percentage,
    "pivot": execute_pivot,
    "rank": execute_rank,
    "rename": execute_rename,
    "replace": execute_replace,
    "replacetext": execute_replacetext,
    "rollup": execute_rollup,
    "select": execute_select,
    "simplify": execute_simplify,
    "sort": execute_sort,
    "split": execute_split,
    "statistics": execute_statistics,
    "substring": execute_substring,
    "text": execute_text,
    "todate": execute_todate,
    "top": execute_top,
    "totals": execute_totals,
    "trim": execute_trim,
    "uniquegroups": execute_uniquegroups,
    "unpivot": execute_unpivot,
    "uppercase": execute_uppercase,
    "waterfall": execute_waterfall,
}



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin

from ..conditions import Condition


class FilterStep(BaseStep):
    name: Literal["filter"] = "filter"
    condition: Condition


class FilterStepWithVariables(FilterStep, StepWithVariablesMixin):
    ...



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class DeleteStep(BaseStep):
    name: Literal["delete"] = "delete"
    columns: list[ColumnName]



from pandas import DataFrame, to_datetime

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ToDateStep


def execute_todate(
    step: ToDateStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    format = step.format
    timestamp_unit = None

    if format is None and df[step.column].dtype == "int64":
        # By default int are understood as timestamps, we prefer to convert it to %Y format if values are < 10_000
        if df[step.column].max() < 10_000:
            format = "%Y"
        else:
            # Timestamps are expected in ms (not in ns, which is pandas' default)
            timestamp_unit = "ms"

    datetime_serie = to_datetime(
        df[step.column], format=format, errors="coerce", unit=timestamp_unit
    )
    return df.assign(**{step.column: datetime_serie})



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ArgminStep


def translate_argmin(step: ArgminStep) -> list[MongoStep]:
    return [
        {
            "$group": {
                "_id": {group: f"${group}" for group in step.groups},
                "_vqbAppArray": {"$push": "$$ROOT"},
                "_vqbAppValueToCompare": {"$min": f"${step.column}"},
            }
        },
        {"$unwind": "$_vqbAppArray"},
        {"$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbAppArray", "$$ROOT"]}}},
        {"$project": {"_vqbAppArray": 0}},
        {
            "$redact": {
                "$cond": [
                    {"$eq": [f"${step.column}", "$_vqbAppValueToCompare"]},
                    "$$KEEP",
                    "$$PRUNE",
                ]
            }
        },
        {"$project": {"_vqbAppValueToCompare": 0}},
    ]



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName

from .utils.combination import (
    PipelineOrDomainName,
    PipelineOrDomainNameOrReference,
    ReferenceResolver,
    resolve_if_reference,
)

JoinColumnsPair = tuple[ColumnName, ColumnName]


class BaseJoinStep(BaseStep):
    name: Literal["join"] = "join"
    type: Literal["left", "inner", "left outer"]
    on: list[JoinColumnsPair] = Field(..., min_items=1)


class JoinStep(BaseJoinStep):
    right_pipeline: PipelineOrDomainName


class JoinStepWithVariable(JoinStep, StepWithVariablesMixin):
    ...


class JoinStepWithRef(BaseJoinStep):
    right_pipeline: PipelineOrDomainNameOrReference

    async def resolve_references(
        self, reference_resolver: ReferenceResolver
    ) -> JoinStepWithVariable | None:
        right_pipeline = await resolve_if_reference(reference_resolver, self.right_pipeline)
        if right_pipeline is None:
            from ..references import ReferenceUnresolved

            raise ReferenceUnresolved()
        return JoinStepWithVariable(
            name=self.name,
            type=self.type,
            on=self.on,
            right_pipeline=right_pipeline,
        )



from pandas import DataFrame, concat

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import AggregateStep, TotalsStep
from weaverbird.pipeline.steps.totals import TotalDimension

from .aggregate import execute_aggregate


def get_total_for_dimension(
    step: TotalsStep, df: DataFrame, total_dimension: TotalDimension, dimensions_to_skip: list
) -> DataFrame:
    # get all group_by columns: all total_dimensions, except the current one + groups
    # all columns that are either not aggregated, or groups, or total will be null
    group_by_columns = step.groups + [
        group_column.total_column
        for group_column in step.total_dimensions
        if group_column != total_dimension
    ]
    aggregations = []
    for aggregation in step.aggregations:
        agg = aggregation.copy()
        agg.columns = agg.new_columns
        aggregations.append(agg)

    aggregation_step = AggregateStep(
        name="aggregate",
        keep_original_granularity=False,
        aggregations=aggregations,
        on=group_by_columns,
    )
    aggregated_df = execute_aggregate(aggregation_step, df)

    aggregated_df[total_dimension.total_column] = total_dimension.total_rows_label
    full_aggregation = aggregated_df.copy()
    # if we are the last dimension, we already computed all combinations
    if total_dimension != step.total_dimensions[-1]:
        for dimension in step.total_dimensions[step.total_dimensions.index(total_dimension) :]:
            if dimension not in dimensions_to_skip and dimension != total_dimension:
                full_aggregation = concat(
                    [
                        full_aggregation,
                        get_total_for_dimension(
                            step, aggregated_df, dimension, dimensions_to_skip + [dimension]
                        ),
                    ]
                )
    return full_aggregation


def get_total_for_dimensions(step: TotalsStep, df: DataFrame) -> DataFrame:
    group_by_columns = step.groups + [dim.total_column for dim in step.total_dimensions]

    aggregation_step = AggregateStep(
        name="aggregate",
        keep_original_granularity=False,
        aggregations=step.aggregations,
        on=group_by_columns,
    )
    aggregated_df = execute_aggregate(aggregation_step, df)

    result_df = DataFrame()
    for dimension in step.total_dimensions:
        result_df = concat([result_df, get_total_for_dimension(step, aggregated_df, dimension, [])])
    return result_df


def execute_totals(
    step: TotalsStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    total_rows = []
    # for total_dimension in step.total_dimensions:
    total_rows.append(get_total_for_dimensions(step, df))

    # rename columns in the base df, so it will match with the total in schema
    col_to_rm = set()
    for aggregation in step.aggregations:
        for col, new_col in zip(aggregation.columns, aggregation.new_columns, strict=True):
            if col != new_col:
                df[new_col] = df[col]
                col_to_rm.add(col)
    for col in col_to_rm:
        del df[col]

    result = concat([df] + total_rows)
    return result



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class TrimStep(BaseStep):
    name: Literal["trim"] = "trim"
    columns: list[ColumnName]



from typing import Any, TypeVar

from pypika import functions
from pypika.dialects import QueryBuilder, SnowflakeQueryBuilder
from pypika.enums import Dialects
from pypika.queries import Query, Selectable
from pypika.terms import Field, Term
from pypika.utils import builder, format_quotes

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateAddWithoutUnderscore,
    DateFormatMapping,
    SQLTranslator,
)

Self = TypeVar("Self", bound="SQLTranslator")


class QuotedSnowflakeQueryBuilder(SnowflakeQueryBuilder):
    QUOTE_CHAR = '"'

    @builder
    def with_(self, selectable: Selectable, name: str) -> "QueryBuilder":
        return super().with_(selectable, format_quotes(name, self.QUOTE_CHAR))


class SnowflakeQuery(Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "QuotedSnowflakeQueryBuilder":
        return QuotedSnowflakeQueryBuilder(**kwargs)


class SnowflakeTranslator(SQLTranslator):
    DIALECT = SQLDialect.SNOWFLAKE
    QUERY_CLS = SnowflakeQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://docs.snowflake.com/en/sql-reference/data-types-datetime#date-and-time-formats
        day_number="DD",
        month_number="MM",
        month_short="MON",
        month_full="MON",
        year="YYYY",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.TO_TIMESTAMP_NTZ

    @classmethod
    def _add_date(
        cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None
    ) -> Term:
        return DateAddWithoutUnderscore(
            date_part=unit.removesuffix("s"), interval=duration, term=target_column
        )

    @classmethod
    def _interval_to_seconds(cls, value: Selectable) -> functions.Function:
        raise NotImplementedError


SQLTranslator.register(SnowflakeTranslator)



import numpy as np
import pandas as pd
from pandas import Series

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import AddMissingDatesStep

# cf. https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
_FREQUENCIES = {"day": "D", "week": "W", "month": "M", "year": "Y"}


def at_begin_period(timestamps: Series, dates_granularity: str):
    # Returns the timezone of the Series, or None in case all objects is naive. In case of mixed
    # objects or non-datetime series, raises an exception
    tz = timestamps.dt.tz
    naive_start_times = timestamps.dt.to_period(_FREQUENCIES[dates_granularity]).dt.start_time
    # Setting the right TZ for all objects
    return naive_start_times.dt.tz_localize(tz)


def execute_addmissingdates(
    step: AddMissingDatesStep,
    df: pd.DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> pd.DataFrame:
    if len(step.groups) > 0:
        groups = df.groupby(step.groups, as_index=False, dropna=False)
    else:
        groups = [("", df)]

    result = pd.DataFrame()
    for key, group in groups:
        # this is used to keep the real date, if it exists, instead of the computed one by pd.Grouper
        group = group.assign(_old_date=group[step.dates_column])

        group_with_missing_dates = group.groupby(
            pd.Grouper(key=step.dates_column, freq=_FREQUENCIES[step.dates_granularity])
        ).agg("first")

        group_with_missing_dates = group_with_missing_dates.reset_index()
        group_with_missing_dates[step.groups] = key

        group_with_missing_dates[step.dates_column] = np.where(
            pd.isna(group_with_missing_dates["_old_date"]),
            at_begin_period(group_with_missing_dates[step.dates_column], step.dates_granularity),
            group_with_missing_dates["_old_date"],
        )
        del group_with_missing_dates["_old_date"]
        # Each group could have duplicate indexes with the previous ones
        # Ignoring the index prevents duplicate index values
        result = pd.concat([result, group_with_missing_dates], ignore_index=True)
    return result



from typing import Literal

from pydantic import validator

from weaverbird.pipeline.steps.utils.base import BaseStep


class SimplifyStep(BaseStep):
    name: Literal["simplify"] = "simplify"
    # All parts of the simplified geometry will be no more than this distance from the original
    tolerance: float = 1.0

    @validator("tolerance")
    def _tolerance_validator(cls, value: float) -> float:
        assert value > 0, "tolerance must be strictly positive"
        return value



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class MovingAverageStep(BaseStep):
    name: Literal["movingaverage"] = "movingaverage"

    value_column: ColumnName | None = None
    column_to_sort: ColumnName
    moving_window: int
    groups: list[ColumnName] = Field(default_factory=list)
    new_column_name: ColumnName | None = None



from typing import Annotated

from pydantic import BaseModel, Field

from .pipeline import PipelineStep, PipelineStepWithVariables, PipelineWithVariables
from .steps import AppendStepWithRefs, DomainStepWithRef, JoinStepWithRef
from .steps.utils.combination import ReferenceResolver

PipelineStepWithRefs = Annotated[
    AppendStepWithRefs | DomainStepWithRef | JoinStepWithRef,
    Field(discriminator="name"),  # noqa: F821
]


class PipelineWithRefs(BaseModel):
    """
    Represents a pipeline in which some steps can reference some other pipelines using the syntax
    `{"type": "ref", "uid": "..."}`
    """

    steps: list[PipelineStepWithRefs | PipelineStepWithVariables | PipelineStep]

    async def resolve_references(
        self, reference_resolver: ReferenceResolver
    ) -> PipelineWithVariables | None:
        """
        Walk the pipeline steps and replace any reference by its corresponding pipeline.
        The sub-pipelines added should also be handled, so that they will be no references anymore in the result.
        """
        resolved_steps: list[PipelineStepWithRefs | PipelineStepWithVariables | PipelineStep] = []
        for step in self.steps:
            resolved_step = (
                await step.resolve_references(reference_resolver)
                if hasattr(step, "resolve_references")
                else step
            )
            if isinstance(resolved_step, PipelineWithVariables):
                resolved_steps.extend(resolved_step.steps)
            elif resolved_step is not None:  # None means the step should be skipped
                resolved_steps.append(resolved_step)

        return PipelineWithVariables(steps=resolved_steps)


class ReferenceUnresolved(Exception):
    """
    Raised when a mandatory reference is not resolved
    """



from pydantic import BaseConfig, Extra
from pydantic import BaseModel as PydanticBaseModel


def to_camelcase(string: str) -> str:
    first_word, *other_words = string.split("_")
    return first_word + "".join(word.capitalize() for word in other_words)


class BaseModel(PydanticBaseModel):
    class Config(BaseConfig):
        allow_population_by_field_name = True
        extra = Extra.forbid
        alias_generator = to_camelcase


class BaseStep(BaseModel):
    name: str

    # None values are excluded, to avoid triggering validations error in front-ends
    def dict(self, *, exclude_none: bool = True, **kwargs) -> dict:
        return super().dict(exclude_none=True, **kwargs)



from typing import Literal

from pydantic import root_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import TemplatedVariable


class RenameStep(BaseStep):
    name: Literal["rename"] = "rename"
    to_rename: list[tuple[str, str]]

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if "oldname" in values and "newname" in values:
            values["to_rename"] = [(values.pop("oldname"), values.pop("newname"))]
        return values


class RenameStepWithVariable(RenameStep, StepWithVariablesMixin):
    to_rename: TemplatedVariable | list[tuple[TemplatedVariable, TemplatedVariable]]
